package com.htfp.weather.griddata.operation;

import com.aliyun.tablestore.grid.GridDataFetcher;
import com.aliyun.tablestore.grid.model.GridDataSet;

import com.htfp.weather.griddata.common.TableConfigBean;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Component;
import ucar.ma2.Array;

import javax.annotation.Resource;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * @Author : shiyi
 * @Date : 2024/1/22 13:53
 * @Description : 根据维度信息获取数据
 */
@Slf4j @Component
@DependsOn({"tableStoreConf","tableConfigBean"})
public class GfsDataFetcher extends BaseTableOperation {
    @Resource
    TableConfigBean tableConfigBean;

    /**
     * 获取指定变量的四维数据(time, lev, lat, lon)
     * @param dataSetId 数据id
     * @param variables 变量名列表
     * @param origin 维度起始索引
     * @param shape 数据形状
     * @return {@link GridDataSet}
     * @throws Exception
     */
    public GridDataSet queryByTableStore(String dataSetId, List<String> variables, int[] origin, int[] shape) throws Exception {
        GridDataFetcher fetcher = tableStoreGrid.getDataFetcher(tableStoreGrid.getDataSetMeta(dataSetId));
        fetcher.setVariablesToGet(variables);
        fetcher.setOriginShape(origin, shape);
        GridDataSet gridDataSet = fetcher.fetch();
        // Grid4D grid4D = fetcher.fetch().getVariable(variable);
        return gridDataSet;
    }

    public GridDataSet getSinglePoint(String dataSetId, List<String> variables, int iTime, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTime, iLev, iLat, iLon};
        int[] shape = new int[] {1, 1, 1, 1};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, variables,
                origin, shape);
        return gridDataSet;
    }

    /**
     *  获取指定变量的二维全平面数据
     * @param variable 变量名
     * @param iLev 高度索引
     * @return {@link Array}
     * @throws Exception
     */
    public Array getPlane(String dataSetId, String variable, int iTime, int iLev) throws Exception {
        int[] origin = new int[] {iTime, iLev, 0, 0};
        int[] shape = new int[] {1, 1, tableConfigBean.latSize, tableConfigBean.lonSize};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable),
                origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     *  获取给定变量的二维平面范围数据
     * @param variable 变量名
     * @param iLev 高度索引
     * @param iLonStart 起始经度索引
     * @param iLatLength 纬度格点数量
     * @param iLonLength 经度格点数量
     * @return {@link Array}
     * @throws Exception
     */
    public Array getPlane(String dataSetId, String variable, int iTime, int iLev, int iLatStart, int iLonStart, int iLatLength, int iLonLength) throws Exception {
        int[] origin = new int[] {iTime, iLev, iLatStart, iLonStart};
        int[] shape = new int[] {1, 1, iLatLength, iLonLength};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable),
                origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     * 获取指定变量在某位置随高度的分布
     * @param variable 变量名
     * @param iLat 纬度索引
     * @param iLon 经度索引
     * @return {@link Array}
     * @throws Exception
     */
    public Array getProfile(String dataSetId, String variable, int iTime, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTime, 0, iLat, iLon};
        int[] shape = new int[] {1, tableConfigBean.levSize, 1, 1};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable),
                origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        return array;
    }

    /**
     * 获取指定变量在某经纬高处的预报序列
     * @param dataSetId 数据集ID
     * @param variable 变量名
     * @param iLev     高度索引
     * @param iLat     纬度索引
     * @param iLon     经度索引
     * @throws Exception
     */
    public Array getSeries(String dataSetId, String variable, int[] iTimeList, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTimeList[0], iLev, iLat, iLon};
        int[] shape = new int[] {iTimeList.length, 1, 1, 1};
        GridDataSet gridDataSet = queryByTableStore(dataSetId, Collections.singletonList(variable), origin, shape);
        Array array = gridDataSet.getVariable(variable).toArray();
        System.out.println("Shape: " + array.shapeToString());
        System.out.println("Data: " + array);
        return array;
    }

    /**
     * 获取指定变量在某位置随时间的变化
     * @param dataSetId 数据集ID
     * @param variableList 变量名列表
     * @param iTImes 时间索引数组
     * @param iLev     高度索引
     * @param iLat     纬度索引
     * @param iLon     经度索引
     * @return
     * @throws Exception
     */
    public GridDataSet getSeries(String dataSetId, List<String> variableList, int[] iTImes, int iLev, int iLat, int iLon) throws Exception {
        int[] origin = new int[] {iTImes[0], iLev, iLat, iLon};
        int[] shape = new int[] {iTImes.length, 1, 1, 1};
        return queryByTableStore(dataSetId, variableList, origin, shape);
    }
}
