package com.htfp.weather.griddata.operation;

import com.aliyun.tablestore.grid.GridDataWriter;
import com.aliyun.tablestore.grid.model.GridDataSetMeta;
import com.aliyun.tablestore.grid.model.StoreOptions;
import com.aliyun.tablestore.grid.model.grid.Grid2D;

import com.htfp.weather.griddata.common.GfsVariableHeightEnum;
import com.htfp.weather.griddata.common.GfsVariableIsobaricEnum;
import com.htfp.weather.griddata.common.TableConfigBean;
import com.htfp.weather.info.Constant;
import com.htfp.weather.utils.MeteoUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.annotation.DependsOn;
import org.springframework.stereotype.Component;
import org.springframework.util.CollectionUtils;
import ucar.ma2.*;
import ucar.nc2.NetcdfFile;
import ucar.nc2.NetcdfFiles;
import ucar.nc2.Variable;

import javax.annotation.PreDestroy;
import javax.annotation.Resource;
import java.io.File;
import java.time.OffsetDateTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

/**
 * @author shi_y
 */
@Slf4j
@Component
@DependsOn("tableConfigBean")
public class GfsDataImport extends BaseTableOperation {
    @Resource
    TableConfigBean tableConfigBean;
    private final ExecutorService executorService = Executors.newFixedThreadPool(5);

    /**
     * init meta data to table store.
     *
     * @param dataSetID
     * @param dataType
     * @param variables
     * @param shape     // * @param attribute meta表的属性列
     * @return
     * @throws Exception
     */
    private GridDataSetMeta initMeta(String dataSetID, DataType dataType, List<String> variables, int[] shape) throws Exception {
        GridDataSetMeta meta = new GridDataSetMeta(
                dataSetID,
                dataType,
                variables,
                shape[0],
                shape[1],
                shape[2],
                shape[3],
                new StoreOptions(StoreOptions.StoreType.SLICE));
        // for (Map.Entry<String, Object> entry : attributes.entrySet()) {
        //     meta.addAttribute(entry.getKey(), entry.getValue());
        // }
        meta.addAttribute("status", "INIT");
        meta.addAttribute("create_time", System.currentTimeMillis());
        tableStoreGrid.putDataSetMeta(meta);
        return meta;
    }

    /**
     * update meta and set status to DONE when data import finished.
     *
     * @param meta
     * @return
     * @throws Exception
     */
    public GridDataSetMeta updateMeta(GridDataSetMeta meta) throws Exception {
        meta.addAttribute("status", "DONE");
        tableStoreGrid.updateDataSetMeta(meta);
        return meta;
    }

    public GridDataSetMeta putMeta(GridDataSetMeta meta) throws Exception {
        meta.addAttribute("status", "DONE");
        tableStoreGrid.putDataSetMeta(meta);
        return meta;
    }


    public void importData(OffsetDateTime refTime) throws Exception {
        log.info("[tablestore] 数据导入开始, refTime = {}...", refTime);
        long start = System.currentTimeMillis();
        List<String> fileList = getFiles(refTime);
        if (CollectionUtils.isEmpty(fileList)) {
            throw new AppExcpetion(ErrorCode.NO_NC_OR_GRIB_FILES);
        }
        String dataSetId = refTime.format(DateTimeFormatter.ofPattern(Constant.DATA_SET_ID_FORMAT_STRING));
        List<String> fileVariables = getFileVariables(tableConfigBean.variableList);
        int[] shape = new int[]{tableConfigBean.timeSizeMax, tableConfigBean.levSize, tableConfigBean.latSize, tableConfigBean.lonSize};
        GridDataSetMeta meta = initMeta(dataSetId, tableConfigBean.dataType, fileVariables, shape);
        List<String> forecastHours = new ArrayList<>(); // 记录到数据库属性中

        // done 2024/5/13: 待优化，用于数据库的索引，必须连续，因此必须保证前一个时刻成功导入后才能导入下一个时刻，数据量大的时候导入时间较长
        // done 2024/5/13: 优化方案，使用多线程，并使用Future来获取结果，但是必须保证每个时刻都成功导入

        List<Future<ImportResult>> futures = new ArrayList<>();
        for (int i = 0; i < fileList.size(); i++) {
            String file = fileList.get(i);
            int iTime = i;
            futures.add(executorService.submit(() -> importFromNcFile(meta, file, iTime)));
            // ImportResult importResult = importFromNcFile(meta, file, iTime);
            // forecastHours.add(String.valueOf(importResult.getForcastHour()));
        }
        for (Future<ImportResult> future : futures) {
            ImportResult importResult = future.get();
            forecastHours.add(String.valueOf(importResult.getForcastHour()));
        }
        long end = System.currentTimeMillis();
        log.info("[tablestore] 数据导入完成, 耗时: {} s, forecastHours: {}", (end - start)/1000., forecastHours);
        meta.setForecastHours(forecastHours);
        meta.addAttribute("reference_time", refTime.toInstant().toEpochMilli());
        putMeta(meta);

    }

    /**
     * read data from netcdf file and write data to table store.
     *
     * @param meta  数据元信息
     * @param file  netcdf文件路径
     * @param iTime 文件序号
     */
    public ImportResult importFromNcFile(GridDataSetMeta meta, String file, int iTime) {
        int time = 0;
        try {
            GridDataWriter writer = tableStoreGrid.getDataWriter(meta);
            NetcdfFile ncFile = NetcdfFiles.open(file);
            // 相对reftime的小时数，文件可能缺失，因此time可能是不连续的，但是iTime是连续的
            time = (int) ncFile.findVariable("time").read().getDouble(0);
            for (String variableName : meta.getVariables()) {
                if ("Wind_speed_gust_surface".equals(variableName) || "Wind_speed_isobaric".equals(variableName) || "Wind_direction_isobaric".equals(variableName)
                    || "Wind_speed_height_above_ground".equals(variableName) || "Wind_direction_height_above_ground".equals(variableName)) {
                    continue;
                }
                Variable variable = ncFile.findVariable(variableName);
                if (variable != null) {
                    int shapeLength = variable.getShape().length;
                    if (shapeLength== 4) {
                        // 气压坐标和高度坐标保持各自的索引
                        int zSize = variable.getShape(1);
                        for (int z = 0; z < zSize; z++) {
                            // 注意高度索引是递增排序
                            Array array = variable.read(new int[]{0, z, 0, 0}, new int[]{1, 1, meta.getxSize(), meta.getySize()});
                            transferUnit(variableName, array);
                            Grid2D grid2D = new Grid2D(array.getDataAsByteBuffer(), variable.getDataType(),
                                    new int[]{0, 0}, new int[]{meta.getxSize(), meta.getySize()});
                            writer.writeGrid2D(variableName, iTime, z, grid2D);
                        }
                    } // TODO 2024/5/23: 导入不含高度坐标的数据
                } else {
                    log.warn("[tablestore] 数据文件 {} 中没有变量 {}", ncFile.getLocation(), variableName);
                }
            }
            // 导入风速风向
            importWind(meta, ncFile, iTime);
            ncFile.close();
        } catch (Exception e) {
            log.error("[tablestore] 导入文件数据失败,_t={}: {}", iTime, file, e);
            return new ImportResult(false, file, time, iTime);
        }
        log.info("[tablestore] 导入文件数据成功,_t={}: {}", iTime, file);
        return new ImportResult(true, file, time, iTime);
    }

    private void transferUnit(String variableName, Array array) {
        if (GfsVariableIsobaricEnum.TEMP.getNameInFile().equals(variableName)) {
            MeteoUtils.kelvin2Celsius(array);
        }
        if (GfsVariableHeightEnum.TEMP.getNameInFile().equals(variableName)) {
            MeteoUtils.kelvin2Celsius(array);
        }
    }

    private void importWind(GridDataSetMeta meta, NetcdfFile ncFile, int iTime) throws Exception {
        // TODO 2024/5/8: 风速风向需要保存到文件中
        for (String suffix : new String[]{"_isobaric", "_height_above_ground"}) {
            Variable uwnd = ncFile.findVariable("u-component_of_wind" + suffix);
            Variable vwnd = ncFile.findVariable("v-component_of_wind" + suffix);
            if (uwnd == null || vwnd == null) {
                return;
            }
            int xsize = meta.getxSize();
            int ysize = meta.getySize();
            int zsize = uwnd.getShape(1);
            GridDataWriter writer = tableStoreGrid.getDataWriter(meta);
            for (int z = 0; z < zsize; z++) {
                Array uwndArray = uwnd.read(new int[]{0, z, 0, 0}, new int[]{1, 1, xsize, ysize});
                Array vwndArray = vwnd.read(new int[]{0, z, 0, 0}, new int[]{1, 1, xsize, ysize});

                Array speedArray = MeteoUtils.calculateWindSpeed(uwndArray, vwndArray);
                Array dirArray = MeteoUtils.calculateWindDirection(uwndArray, vwndArray);
                Grid2D speedGrid2D = new Grid2D(speedArray.getDataAsByteBuffer(), speedArray.getDataType(),
                        new int[]{0, 0}, new int[]{xsize, ysize});
                writer.writeGrid2D("Wind_speed" + suffix, iTime, z, speedGrid2D);

                Grid2D dirGrid2D = new Grid2D(dirArray.getDataAsByteBuffer(), dirArray.getDataType(),
                        new int[]{0, 0}, new int[]{xsize, ysize});
                writer.writeGrid2D("Wind_direction" + suffix, iTime, z, dirGrid2D);
            }
        }
    }

    private List<String> getFileVariables(List<String> variables) {
        List<String> fileVariables = new ArrayList<>();
        for (String variable : variables) {
            String variableIsobaric = GfsVariableIsobaricEnum.getGfsVariableName(variable);
            if (variableIsobaric != null) {
                fileVariables.add(variableIsobaric);
            } else {
                log.warn("GFS数据文件中（气压坐标）没有 {} 对应的变量", variable);
            }

            String variableAboveGround = GfsVariableHeightEnum.getGfsVariableName(variable);
            if (variableAboveGround != null) {
                fileVariables.add(variableAboveGround);
            } else {
                log.warn("GFS数据文件中（高度坐标）没有 {} 对应的变量", variable);
            }
        }
        return fileVariables;
    }

    /**
     * 获取文件目录中的所有文件绝对路径
     *
     * @param refTime，UTC起报时间
     * @return
     */
    private List<String> getFiles(OffsetDateTime refTime) {
        String dataFolder = refTime.format(DateTimeFormatter.ofPattern(Constant.DATA_FOLDER_STRING));
        File fileDir = new File(tableConfigBean.dataDir + dataFolder);
        if (!fileDir.exists()) {
            log.warn("文件夹 {} 不存在", fileDir);
            return null;
        }
        List<String> fileList = new ArrayList<>();
        String[] files = fileDir.list((dir, name) -> {
            return name.startsWith("BJT") && name.endsWith("grib2"); // 判断文件名是否以“BJT”开头
        });
        if (files == null || files.length == 0) {
            log.warn("文件夹 {} 中没有符合条件的文件", fileDir);
            return fileList;
        }
        for (String filename : files) {
            fileList.add(fileDir.getAbsolutePath() + File.separator + filename);
        }
        return fileList;
    }

    @PreDestroy
    public void end() {
        close();
    }

    @Data
    public static class ImportResult {
        private boolean success;
        private String file;
        private int forcastHour;
        private int iTime;

        public ImportResult(boolean success, String file, int forcastHour, int iTime) {
            this.success = success;
            this.file = file;
            this.forcastHour = forcastHour;
            this.iTime = iTime;
        }
    }

}
