package com.htfp.weather.web.service.surfaceapi;

import com.htfp.weather.info.CaiYunInfo;
import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.MeteoUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.caiyun.*;
import com.htfp.weather.web.pojo.caiyun.CaiYunNowResponse.NowResult.RealTime;
import com.htfp.weather.web.pojo.caiyun.CaiYunForecastResponse.HourlyResult.Hourly;
import com.htfp.weather.web.pojo.caiyun.CaiYunWarningResponse.AlertResult.Alert.AlertContent;
import com.htfp.weather.web.pojo.response.NowWeatherStatus;
import com.htfp.weather.web.pojo.response.SurfaceWeatherWarning;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import static com.htfp.weather.utils.HttpClientUtils.sendGet;

/**
 * @Author : shiyi
 * @Date : 2024/4/15 18:42
 * @Description : 彩云天气
 */
@Service("caiyun") @Slf4j
public class CaiYunServiceImpl implements ISurfaceDataService {

    private final String key = "Tc9tgOYr5jlPPlEw";
    private final HashMap<String, String> variableNameMap =
            new HashMap<String, String>() {{
            }};
    static final String STATUS_OK = "ok";
    public <T extends CaiYunBaseResponse> T caiYunRequest(String url, HashMap<String, String> params, Class<T> responseClass) {
        params.put("units", "metric:v2");
        try {
            T response = sendGet(url, params, responseClass);
            if (response != null && STATUS_OK.equals(response.getStatus())) {
                return response;
            } else if (response == null) {
                log.error("[彩云天气] 请求失败");
                throw new AppExcpetion(ErrorCode.CAI_YUN_REQUEST_ERROR);
            } else {
                log.error("[彩云天气] {}: {}", ErrorCode.CAI_YUN_REQUEST_ERROR.getMsg(), response);
                throw new AppExcpetion(ErrorCode.CAI_YUN_REQUEST_ERROR);
            }
        } catch (Exception e) {
            log.error("[彩云天气] 请求结果处理错误, {}", e.getMessage());
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取地面指定位置的天气状况
     *
     * @throws Exception
     */
    @Override
    public NowWeatherStatus getNowSurfaceWeatherStatus(double lat, double lon) throws Exception {
        CaiYunNowResponse caiYunNowWeather = getCaiYunNowWeather(lat, lon);
        return buildNowSurfaceWeatherStatus(caiYunNowWeather);
    }


    /**
     * 获取地面指定位置某变量的24小时预报序列
     *
     * @throws Exception
     */
    @Override
    public TimeSeriesDataset getForecastSeries(double lat, double lon) throws Exception {
        Hourly hourly = getCaiYunForecast24(lat, lon);
        return buildForecastSeries(hourly);
    }

    /**
     * 获取地面指定位置的气象预警信息
     *
     * @throws Exception
     */
    @Override
    public List<SurfaceWeatherWarning> getSurfaceWarning(double lat, double lon) throws Exception {
        List<AlertContent> caiYunWarnings = getCaiYunWarning(lat, lon);
        return buildSurfaceWarning(caiYunWarnings);
    }

    public CaiYunNowResponse getCaiYunNowWeather(double lat, double lon) throws Exception {
        String url = String.format("https://api.caiyunapp.com/v2.6/%s/%.2f,%.2f/realtime", key, lon, lat);
        HashMap<String, String> params = new HashMap<>();
        return caiYunRequest(url, params, CaiYunNowResponse.class);
    }

    private Hourly getCaiYunForecast24(double lat, double lon) {
        String url = String.format("https://api.caiyunapp.com/v2.6/%s/%.2f,%.2f/hourly", key, lon, lat);
        HashMap<String, String> params = new HashMap<>();
        params.put("hourlysteps", "24");
        CaiYunForecastResponse response = caiYunRequest(url, params, CaiYunForecastResponse.class);
        return response.getHourlyResult().getHourly();
    }


    public List<AlertContent> getCaiYunWarning(double lat, double lon) throws Exception {
        String url = String.format("https://api.caiyunapp.com/v2.6/%s/%.2f,%.2f/realtime", key, lon, lat);
        HashMap<String, String> params = new HashMap<>();
        params.put("alert", "true");
        CaiYunWarningResponse response = caiYunRequest(url, params, CaiYunWarningResponse.class);
        return response.getResult().getAlert().getContent();
    }

    private NowWeatherStatus buildNowSurfaceWeatherStatus(CaiYunNowResponse nowResponse) {
        NowWeatherStatus nowWeatherStatus = new NowWeatherStatus();
        try {
            RealTime realtime = nowResponse.getNowResult().getRealtime();
            String time = DateTimeUtils.getTimeStringFromMilli(nowResponse.getServerTime()*1000, nowResponse.getTimeZone());
            nowWeatherStatus.setTime(time);
            nowWeatherStatus.setWeatherText(CaiYunInfo.getSkyConName(realtime.getSkycon()));
            nowWeatherStatus.setTemp(realtime.getTemperature());
            nowWeatherStatus.setWindSpeed((float) MeteoUtils.kmPerHour2mPerSecond(realtime.getWind().getSpeed()));
            nowWeatherStatus.setWind360(realtime.getWind().getDirection());
            nowWeatherStatus.setHumidity(realtime.getHumidity());
            nowWeatherStatus.setCloud(realtime.getCloudrate());

            CaiYunNowResponse.NowResult.Precipitation.Local localPrecipitation = realtime.getPrecipitation().getLocal();
            if ("ok".equals(localPrecipitation.getStatus())) {
                nowWeatherStatus.setPrecip(localPrecipitation.getIntensity());
            }
            return nowWeatherStatus;
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.CAI_YUN_DATA_PARSE_ERROR);
        }
    }

    private TimeSeriesDataset buildForecastSeries(Hourly hourly) {
        try {
            TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(24);
            for (int i = 0; i < 24; i++) {
                timeSeriesDataset.getTime().add(hourly.getTemperature().get(i).getDatetime());
                timeSeriesDataset.getTemp()[i] = hourly.getTemperature().get(i).getValue();
                timeSeriesDataset.getWindSpeed()[i] = (float) MeteoUtils.kmPerHour2mPerSecond(hourly.getWind().get(i).getSpeed());
                timeSeriesDataset.getWind360()[i] = hourly.getWind().get(i).getDirection();
                timeSeriesDataset.getHumidity()[i] = hourly.getHumidity().get(i).getValue();
                timeSeriesDataset.getCloud()[i] = hourly.getCloudrate().get(i).getValue();
                timeSeriesDataset.getPrecip()[i] = hourly.getPrecipitation().get(i).getValue();
            }
            return timeSeriesDataset;
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.CAI_YUN_DATA_PARSE_ERROR, e.getMessage());
        }
    }

    private List<SurfaceWeatherWarning> buildSurfaceWarning(List<AlertContent> warnings) {
        try {
            List<SurfaceWeatherWarning> surfaceWeatherWarnings = new ArrayList<>();
            for (AlertContent warning : warnings) {
                SurfaceWeatherWarning surfaceWeatherWarning = new SurfaceWeatherWarning();
                surfaceWeatherWarning.setPubTime(DateTimeUtils.getTimeStringFromMilli(warning.getPubtimestamp()));
                surfaceWeatherWarning.setTitle(warning.getTitle());
                surfaceWeatherWarning.setText(warning.getDescription());
                surfaceWeatherWarning.setTypeName(CaiYunInfo.getAlertName(warning.getCode()));
                surfaceWeatherWarning.setSeverityColor(CaiYunInfo.getAlertColor(warning.getCode()));
                surfaceWeatherWarning.setSource(warning.getSource());
                surfaceWeatherWarnings.add(surfaceWeatherWarning);
            }
            return surfaceWeatherWarnings;
        } catch (Exception e) {
            throw new AppExcpetion(ErrorCode.CAI_YUN_DATA_PARSE_ERROR);
        }
    }
}
