package com.htfp.weather.web.service.surfaceapi;

import com.htfp.weather.utils.DateTimeUtils;
import com.htfp.weather.utils.MeteoUtils;
import com.htfp.weather.web.exception.AppExcpetion;
import com.htfp.weather.web.exception.ErrorCode;
import com.htfp.weather.web.pojo.response.SurfaceWeatherWarning;
import com.htfp.weather.web.pojo.response.TimeSeriesDataset;
import com.htfp.weather.web.pojo.response.NowWeatherStatus;
import com.htfp.weather.web.pojo.hefeng.*;
import com.htfp.weather.web.pojo.hefeng.HeFengNowResponse.HeFengNow;
import com.htfp.weather.web.pojo.hefeng.HeFengForecastResponse.HeFengForecastHour;
import com.htfp.weather.web.pojo.hefeng.HeFengWarningResponse.HeFengWarning;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.MessageDigest;
import java.util.*;

import static com.htfp.weather.utils.HttpClientUtils.*;

/**
 * @Author : shiyi
 * @Date : 2024/1/23 10:57
 * @Description : 和风天气服务
 */
@Service("hefeng")
@Slf4j
public class HeFengServiceImpl implements ISurfaceDataService {
    @Value("${hefeng.key}")
    private String key;
    private final HashMap<String, String> variableNameMap =
            new HashMap<String, String>() {{
                put("temp", "temp");
                put("windSpeed", "windSpeed");
                put("windDir", "windDir");
                // put("windScale", "windScale");
                put("humidity", "humidity");
                put("cloud", "cloud");
                put("pressure", "pressure");
                put("precip", "precip");
            }};
    static final String STATUS_OK = "200";
    static final String STATUS_NO_DATA = "204";

    public <T extends HeFengBaseResponse> T heFengRequest(String url, HashMap<String, String> params, Class<T> responseClass) throws Exception {
        // String signature = getSignature(params, key);
        // HashMap<String, String> params = new HashMap<>();
        // params.put("sign", signature);
        params.put("key", key);
        try {
            T response = sendGet(url, params, responseClass);
            if (response != null && STATUS_OK.equals(response.getCode())) {
                return response;
            } else if (response == null) {
                log.error("[和风天气] 请求失败");
                throw new AppExcpetion(ErrorCode.HE_FENG_REQUEST_ERROR);
            } else if (STATUS_NO_DATA.equals(response.getCode())) {
                log.error("[和风天气] {}: {}", ErrorCode.HE_FENG_THIS_AREA_HAVE_NO_DATA.getMsg(), response);
                throw new AppExcpetion(ErrorCode.HE_FENG_THIS_AREA_HAVE_NO_DATA);
            } else {
                log.error("[和风天气] {}: {}", ErrorCode.HE_FENG_THIS_AREA_HAVE_NO_DATA.getMsg(), response);
                throw new AppExcpetion(ErrorCode.HE_FENG_THIS_AREA_HAVE_NO_DATA);
            }
        } catch (Exception e) {
            log.error("[和风天气] 请求结果处理错误, {}", e.getMessage());
            throw new RuntimeException(e);
        }
    }

    /**
     * 获取地面指定位置的天气状况
     *
     * @throws Exception
     */
    @Override
    public NowWeatherStatus getNowSurfaceWeatherStatus(double lat, double lon) throws Exception {
        HeFengNow nowWeather = getFeFengNowWeather(lat, lon);
        return buildNowSurfaceWeatherStatus(nowWeather);
    }

    /**
     * 获取地面指定位置某变量的24小时预报序列
     *
     * @throws Exception
     */
    @Override
    public TimeSeriesDataset getForecastSeries(double lat, double lon) throws Exception {
        List<HeFengForecastHour> hourly = getHeFengForecast24(lat, lon);
        return buildForecastSeries(hourly);
    }

    @Override
    public List<SurfaceWeatherWarning> getSurfaceWarning(double lat, double lon) throws Exception {
        List<HeFengWarning> heFengWarnings = getHeFengWarning(lat, lon);
        return buildSurfaceWarning(heFengWarnings);
    }

    public HeFengNow getFeFengNowWeather(double lat, double lon) throws Exception {
        String url = "https://devapi.qweather.com/v7/grid-weather/now";
        String location = String.format("%.2f,%.2f", lon, lat);
        HashMap<String, String> params = new HashMap<>();
        params.put("location", location);

        HeFengNowResponse response = heFengRequest(url, params, HeFengNowResponse.class);
        return response.getNow();
    }

    public List<HeFengForecastHour> getHeFengForecast24(double lat, double lon) throws Exception {
        String url = "https://devapi.qweather.com/v7/grid-weather/24h";
        String location = String.format("%.2f,%.2f", lon, lat);
        HashMap<String, String> params = new HashMap<>();
        params.put("location", location);

        HeFengForecastResponse response = heFengRequest(url, params, HeFengForecastResponse.class);
        return response.getHourly();

    }

    public List<HeFengWarning> getHeFengWarning(double lat, double lon) throws Exception {
        String url = "https://devapi.qweather.com/v7/warning/now";
        String location = String.format("%.2f,%.2f", lon, lat);
        HashMap<String, String> params = new HashMap<>();
        params.put("location", location);
        HeFengWarningResponse response = heFengRequest(url, params, HeFengWarningResponse.class);
        return response.getWarning();
    }

    private TimeSeriesDataset buildForecastSeries(List<HeFengForecastHour> hourly) {
        TimeSeriesDataset timeSeriesDataset = new TimeSeriesDataset(hourly.size());
        for (int i = 0; i < hourly.size(); i++) {
            HeFengForecastHour heFengForecastHour  = hourly.get(i);
            processHeFengData(heFengForecastHour);
            timeSeriesDataset.getTime().add(heFengForecastHour.getFxTime());
            timeSeriesDataset.getTemp()[i] = heFengForecastHour.getTemp();
            timeSeriesDataset.getWindSpeed()[i] = heFengForecastHour.getWindSpeed();
            timeSeriesDataset.getWind360()[i] = heFengForecastHour.getWind360();
            // timeSeriesDataset.getWindScale()[i] = heFengForecastHour.getWindScale();
            timeSeriesDataset.getHumidity()[i] = heFengForecastHour.getHumidity();
            // timeSeriesDataset.getPressure()[i] = heFengForecastHour.getPressure();
            timeSeriesDataset.getCloud()[i] = heFengForecastHour.getCloud();
            timeSeriesDataset.getPrecip()[i] = heFengForecastHour.getPrecip();
        }
        return timeSeriesDataset;
    }

    /**
     * 从和风返回的数据中解析指定变量序列
     */
    // private TimeSeries<Float> buildForecastSeries(List<HeFengForecastHour> hourly, String variableName) throws IllegalAccessException, NoSuchFieldException {
    //     TimeSeries<Float> timeSeries = new TimeSeries<>();
    //     List<String> time = new ArrayList<>();
    //     List<Float> variable = new ArrayList<>();
    //     String heFengName = variableNameMap.get(variableName);
    //     if (heFengName == null) {
    //         log.warn("[和风天气] 未找到{}对应的变量名", variableName);
    //         throw new AppExcpetion("[和风天气] 未找到" + variableName + "对应的变量名");
    //     }
    //     Field field = BaseHeFengData.class.getDeclaredField(heFengName);
    //     field.setAccessible(true);
    //     for (HeFengForecastHour heFengForecastHour : hourly) {
    //         processHeFengData(heFengForecastHour);
    //         time.add(heFengForecastHour.getFxTime());
    //         variable.add((Float) field.get(heFengForecastHour));
    //     }
    //     timeSeries.setName(variableName);
    //     timeSeries.setTime(time);
    //     timeSeries.setData(variable);
    //     return timeSeries;
    // }
    private NowWeatherStatus buildNowSurfaceWeatherStatus(HeFengNow now) {
        NowWeatherStatus nowWeatherStatus = new NowWeatherStatus();
        processHeFengData(now);
        nowWeatherStatus.setTime(now.getObsTime());
        nowWeatherStatus.setWeatherText(now.getText());
        nowWeatherStatus.setTemp(now.getTemp());
        nowWeatherStatus.setWindSpeed(now.getWindSpeed());
        nowWeatherStatus.setWind360(now.getWind360());
        // nowSurfaceWeatherStatus.setWindDir(now.getWindDir());
        // nowSurfaceWeatherStatus.setWindScale(now.getWindScale());
        nowWeatherStatus.setHumidity(now.getHumidity());
        // nowWeatherStatus.setPressure(now.getPressure());
        nowWeatherStatus.setCloud(now.getCloud());
        nowWeatherStatus.setPrecip(now.getPrecip());
        return nowWeatherStatus;
    }

    private List<SurfaceWeatherWarning> buildSurfaceWarning(List<HeFengWarning> heFengWarnings) {
        List<SurfaceWeatherWarning> surfaceWeatherWarnings = new ArrayList<>();
        for (HeFengWarning heFengWarning : heFengWarnings) {
            SurfaceWeatherWarning surfaceWeatherWarning = new SurfaceWeatherWarning();
            surfaceWeatherWarning.setPubTime(heFengWarning.getPubTime());
            // surfaceWeatherWarning.setStartTime(heFengWarning.getStartTime());
            surfaceWeatherWarning.setTitle(heFengWarning.getTitle());
            surfaceWeatherWarning.setText(heFengWarning.getText());
            surfaceWeatherWarning.setTypeName(heFengWarning.getTypeName());
            surfaceWeatherWarning.setSeverityColor(heFengWarning.getSeverityColor());
            surfaceWeatherWarning.setSource(heFengWarning.getSender());
            surfaceWeatherWarnings.add(surfaceWeatherWarning);
        }
        return surfaceWeatherWarnings;
    }

    /**
     * 从和风返回的数据中对时间进行处理，风速单位处理
     *
     * @param data
     * @return
     */
    private BaseHeFengData processHeFengData(BaseHeFengData data) {
        String format = "yyyy-MM-dd'T'HH:mmxxx";
        if (data instanceof HeFengForecastHour) {
            String fxTime = DateTimeUtils.getBJTDateTimeStringFromUTC(((HeFengForecastHour) data).getFxTime(), format);
            ((HeFengForecastHour) data).setFxTime(fxTime);
        } else if (data instanceof HeFengNow) {
            String obsTime = DateTimeUtils.getBJTDateTimeStringFromUTC(((HeFengNow) data).getObsTime(), format);
            ((HeFengNow) data).setObsTime(obsTime);
        }
        float windSpeed = (float) MeteoUtils.kmPerHour2mPerSecond(data.getWindSpeed());
        data.setWindSpeed(windSpeed);
        return data;
    }

    /**
     * 和风天气签名生成算法-JAVA版本
     *
     * @param params 请求参数集，所有参数必须已转换为字符串类型
     * @param secret 签名密钥（用户的认证key）
     * @return 签名
     * @throws IOException
     */
    public String getSignature(HashMap<String, String> params, String secret) throws Exception {
        // 先将参数以其参数名的字典序升序进行排序
        Map<String, String> sortedParams = new TreeMap<>(params);
        Set<Map.Entry<String, String>> entrys = sortedParams.entrySet();

        // 遍历排序后的字典，将所有参数按"key=value"格式拼接在一起
        StringBuilder baseString = new StringBuilder();
        for (Map.Entry<String, String> param : entrys) {
            // sign参数 和 空值参数 不加入算法
            if (param.getValue() != null && !"".equals(param.getKey().trim()) && !"sign".equals(param.getKey().trim()) && !"key".equals(param.getKey().trim()) && !"".equals(param.getValue().trim())) {
                baseString.append(param.getKey().trim()).append("=").append(param.getValue().trim()).append("&");
            }
        }
        if (baseString.length() > 0) {
            baseString.deleteCharAt(baseString.length() - 1).append(secret);
        }
        // 使用MD5对待签名串求签
        try {
            MessageDigest md5 = MessageDigest.getInstance("MD5");
            byte[] bytes = md5.digest(baseString.toString().getBytes(StandardCharsets.UTF_8));
            return new String(encodeHex(bytes));
        } catch (GeneralSecurityException ex) {
            throw ex;
        }
    }

    public char[] encodeHex(byte[] data) {
        int l = data.length;
        char[] out = new char[l << 1];
        int i = 0;
        char[] toDigits = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f'};
        for (int var5 = 0; i < l; ++i) {
            out[var5++] = toDigits[(240 & data[i]) >>> 4];
            out[var5++] = toDigits[15 & data[i]];
        }
        return out;
    }
}
